@@ -40,11 +40,13 @@
 #include <sys/stat.h>
 #include <sys/wait.h>
 #include <sys/socket.h>
+#include <sys/un.h>
 #include <sys/types.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
 #include <ctype.h>
 #include <errno.h>
+#include <locale.h>
 
 #include <NetworkManager.h>
 #include <NetworkManagerVPN.h>
@@ -53,6 +55,7 @@
 #include "nm-openvpn-service.h"
 #include "nm-utils.h"
 #include "common/utils.h"
+#include "common/nm-glib-compat.h"
 
 #if !defined(DIST_VERSION)
 # define DIST_VERSION VERSION
@@ -68,11 +71,13 @@ G_DEFINE_TYPE (NMOpenvpnPlugin, nm_openvpn_plugin, NM_TYPE_VPN_PLUGIN)
 #define NM_OPENVPN_PLUGIN_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), NM_TYPE_OPENVPN_PLUGIN, NMOpenvpnPluginPrivate))
 
 typedef struct {
+	char *default_username;
 	char *username;
 	char *password;
 	char *priv_key_pass;
 	char *proxy_username;
 	char *proxy_password;
+	char *pending_auth;
 	GIOChannel *socket_channel;
 	guint socket_channel_eventid;
 } NMOpenvpnPluginIOData;
@@ -82,6 +87,8 @@ typedef struct {
 	guint connect_timer;
 	guint connect_count;
 	NMOpenvpnPluginIOData *io_data;
+	gboolean interactive;
+	char *mgt_path;
 } NMOpenvpnPluginPrivate;
 
 typedef struct {
@@ -97,6 +104,7 @@ static ValidProperty valid_properties[] = {
 	{ NM_OPENVPN_KEY_CA,                   G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_CERT,                 G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_CIPHER,               G_TYPE_STRING, 0, 0, FALSE },
+	{ NM_OPENVPN_KEY_KEYSIZE,              G_TYPE_INT, 1, 65535, FALSE },
 	{ NM_OPENVPN_KEY_COMP_LZO,             G_TYPE_BOOLEAN, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_CONNECTION_TYPE,      G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_FRAGMENT_SIZE,        G_TYPE_INT, 0, G_MAXINT, FALSE },
@@ -119,7 +127,10 @@ static ValidProperty valid_properties[] = {
 	{ NM_OPENVPN_KEY_TA,                   G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_TA_DIR,               G_TYPE_INT, 0, 1, FALSE },
 	{ NM_OPENVPN_KEY_TAP_DEV,              G_TYPE_BOOLEAN, 0, 0, FALSE },
-	{ NM_OPENVPN_KEY_TLS_REMOTE,	       G_TYPE_STRING, 0, 0, FALSE },
+	{ NM_OPENVPN_KEY_DEV,                  G_TYPE_STRING, 0, 0, FALSE },
+	{ NM_OPENVPN_KEY_DEV_TYPE,             G_TYPE_STRING, 0, 0, FALSE },
+	{ NM_OPENVPN_KEY_TLS_REMOTE,           G_TYPE_STRING, 0, 0, FALSE },
+	{ NM_OPENVPN_KEY_REMOTE_CERT_TLS,      G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_TUNNEL_MTU,           G_TYPE_INT, 0, G_MAXINT, FALSE },
 	{ NM_OPENVPN_KEY_USERNAME,             G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_PASSWORD"-flags",     G_TYPE_STRING, 0, 0, FALSE },
@@ -266,17 +277,8 @@ nm_openvpn_secrets_validate (NMSettingVPN *s_vpn, GError **error)
 	ValidateInfo info = { &valid_secrets[0], &validate_error, FALSE };
 
 	nm_setting_vpn_foreach_secret (s_vpn, validate_one_property, &info);
-	if (!info.have_items) {
-		g_set_error (error,
-		             NM_VPN_PLUGIN_ERROR,
-		             NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
-		             "%s",
-		             _("No VPN secrets!"));
-		return FALSE;
-	}
-
 	if (validate_error) {
-		*error = validate_error;
+		g_propagate_error (error, validate_error);
 		return FALSE;
 	}
 	return TRUE;
@@ -302,6 +304,7 @@ nm_openvpn_disconnect_management_socket (NMOpenvpnPlugin *plugin)
 
 	g_free (io_data->username);
 	g_free (io_data->proxy_username);
+	g_free (io_data->pending_auth);
 
 	if (io_data->password)
 		memset (io_data->password, 0, strlen (io_data->password));
@@ -395,15 +398,105 @@ write_user_pass (GIOChannel *channel,
 	g_free (buf);
 }
 
+static gboolean
+handle_auth (NMOpenvpnPluginIOData *io_data,
+             const char *requested_auth,
+             const char **out_message,
+             char ***out_hints)
+{
+	gboolean handled = FALSE;
+	guint i = 0;
+	char **hints = NULL;
+
+	g_return_val_if_fail (requested_auth != NULL, FALSE);
+	g_return_val_if_fail (out_message != NULL, FALSE);
+	g_return_val_if_fail (out_hints != NULL, FALSE);
+
+	if (strcmp (requested_auth, "Auth") == 0) {
+		const char *username = io_data->username;
+
+		/* Fall back to the default username if it wasn't overridden by the user */
+		if (!username)
+			username = io_data->default_username;
+
+		if (username != NULL && io_data->password != NULL) {
+			write_user_pass (io_data->socket_channel,
+			                 requested_auth,
+			                 username,
+			                 io_data->password);
+		} else {
+			hints = g_new0 (char *, 3);
+			if (!username) {
+				hints[i++] = NM_OPENVPN_KEY_USERNAME;
+				*out_message = _("A username is required.");
+			}
+			if (!io_data->password) {
+				hints[i++] = NM_OPENVPN_KEY_PASSWORD;
+				*out_message = _("A password is required.");
+			}
+			if (!username && !io_data->password)
+				*out_message = _("A username and password are required.");
+		}
+		handled = TRUE;
+	} else if (!strcmp (requested_auth, "Private Key")) {
+		if (io_data->priv_key_pass) {
+			char *qpass, *buf;
+
+			/* Quote strings passed back to openvpn */
+			qpass = ovpn_quote_string (io_data->priv_key_pass);
+			buf = g_strdup_printf ("password \"%s\" \"%s\"\n", requested_auth, qpass);
+			memset (qpass, 0, strlen (qpass));
+			g_free (qpass);
+
+			/* Will always write everything in blocking channels (on success) */
+			g_io_channel_write_chars (io_data->socket_channel, buf, strlen (buf), NULL, NULL);
+			g_io_channel_flush (io_data->socket_channel, NULL);
+			g_free (buf);
+		} else {
+			hints = g_new0 (char *, 2);
+			hints[i++] = NM_OPENVPN_KEY_CERTPASS;
+			*out_message = _("A private key password is required.");
+		}
+		handled = TRUE;
+	} else if (strcmp (requested_auth, "HTTP Proxy") == 0) {
+		if (io_data->proxy_username != NULL && io_data->proxy_password != NULL) {
+			write_user_pass (io_data->socket_channel,
+			                 requested_auth,
+			                 io_data->proxy_username,
+			                 io_data->proxy_password);
+		} else {
+			hints = g_new0 (char *, 3);
+			if (!io_data->proxy_username) {
+				hints[i++] = NM_OPENVPN_KEY_HTTP_PROXY_USERNAME;
+				*out_message = _("An HTTP Proxy username is required.");
+			}
+			if (!io_data->proxy_password) {
+				hints[i++] = NM_OPENVPN_KEY_HTTP_PROXY_PASSWORD;
+				*out_message = _("An HTTP Proxy password is required.");
+			}
+			if (!io_data->proxy_username && !io_data->proxy_password)
+				*out_message = _("An HTTP Proxy username and password are required.");
+		}
+		handled = TRUE;
+	}
+
+	*out_hints = hints;
+	return handled;
+}
+
 static gboolean
 handle_management_socket (NMVPNPlugin *plugin,
                           GIOChannel *source,
                           GIOCondition condition,
                           NMVPNPluginFailure *out_failure)
 {
-	NMOpenvpnPluginIOData *io_data = NM_OPENVPN_PLUGIN_GET_PRIVATE (plugin)->io_data;
+	NMOpenvpnPluginPrivate *priv = NM_OPENVPN_PLUGIN_GET_PRIVATE (plugin);
 	gboolean again = TRUE;
-	char *str = NULL, *auth = NULL, *buf;
+	char *str = NULL, *auth = NULL;
+	const char *message = NULL;
+	char **hints = NULL;
+
+	g_assert (out_failure);
 
 	if (!(condition & G_IO_IN))
 		return TRUE;
@@ -411,60 +504,73 @@ handle_management_socket (NMVPNPlugin *plugin,
 	if (g_io_channel_read_line (source, &str, NULL, NULL, NULL) != G_IO_STATUS_NORMAL)
 		return TRUE;
 
-	if (strlen (str) < 1)
-		goto out;
+	if (!str[0]) {
+		g_free (str);
+		return TRUE;
+	}
+
+	if (debug)
+		g_message ("VPN request '%s'", str);
 
 	auth = get_detail (str, ">PASSWORD:Need '");
 	if (auth) {
-		if (strcmp (auth, "Auth") == 0) {
-			if (io_data->username != NULL && io_data->password != NULL)
-				write_user_pass (source, auth, io_data->username, io_data->password);
-			else
-				g_warning ("Auth requested but one of username or password is missing");
-		} else if (!strcmp (auth, "Private Key")) {
-			if (io_data->priv_key_pass) {
-				char *qpass;
+		if (priv->io_data->pending_auth)
+			g_free (priv->io_data->pending_auth);
+		priv->io_data->pending_auth = auth;
 
-				/* Quote strings passed back to openvpn */
-				qpass = ovpn_quote_string (io_data->priv_key_pass);
-				buf = g_strdup_printf ("password \"%s\" \"%s\"\n", auth, qpass);
-				memset (qpass, 0, strlen (qpass));
-				g_free (qpass);
-
-				/* Will always write everything in blocking channels (on success) */
-				g_io_channel_write_chars (source, buf, strlen (buf), NULL, NULL);
-				g_io_channel_flush (source, NULL);
-				g_free (buf);
-			} else
-				g_warning ("Certificate password requested but private key password == NULL");
-		} else if (strcmp (auth, "HTTP Proxy") == 0) {
-			if (io_data->proxy_username != NULL && io_data->proxy_password != NULL)
-				write_user_pass (source, auth, io_data->proxy_username, io_data->proxy_password);
-			else
-				g_warning ("HTTP Proxy auth requested but either proxy username or password is missing");
+		if (handle_auth (priv->io_data, auth, &message, &hints)) {
+			/* Request new secrets if we need any */
+			if (message) {
+				if (priv->interactive) {
+					if (debug) {
+						char *joined = hints ? g_strjoinv (",", (char **) hints) : "none";
+						g_message ("Requesting new secrets: '%s' (%s)", message, joined);
+						g_free (joined);
+					}
+					nm_vpn_plugin_secrets_required (plugin, message, (const char **) hints);
+				} else {
+					/* Interactive not allowed, can't ask for more secrets */
+					g_warning ("More secrets required but cannot ask interactively");
+					*out_failure = NM_VPN_PLUGIN_FAILURE_LOGIN_FAILED;
+					again = FALSE;
+				}
+			}
+			if (hints)
+				g_free (hints);  /* elements are 'const' */
 		} else {
-			g_warning ("No clue what to send for username/password request for '%s'", auth);
-			if (out_failure)
-				*out_failure = NM_VPN_PLUGIN_FAILURE_CONNECT_FAILED;
+			g_warning ("Unhandled management socket request '%s'", auth);
+			*out_failure = NM_VPN_PLUGIN_FAILURE_CONNECT_FAILED;
 			again = FALSE;
 		}
-		g_free (auth);
+		goto out;
 	}
 
 	auth = get_detail (str, ">PASSWORD:Verification Failed: '");
 	if (auth) {
-		if (!strcmp (auth, "Auth"))
+		gboolean fail = TRUE;
+
+		if (!strcmp (auth, "Auth")) {
 			g_warning ("Password verification failed");
-		else if (!strcmp (auth, "Private Key"))
+			if (priv->interactive) {
+				/* Clear existing password in interactive mode, openvpn
+				 * will request a new one after restarting.
+				 */
+				if (priv->io_data->password)
+					memset (priv->io_data->password, 0, strlen (priv->io_data->password));
+				g_clear_pointer (&priv->io_data->password, g_free);
+				fail = FALSE;
+			}
+		} else if (!strcmp (auth, "Private Key"))
 			g_warning ("Private key verification failed");
 		else
 			g_warning ("Unknown verification failed: %s", auth);
 
-		g_free (auth);
-
-		if (out_failure)
+		if (fail) {
 			*out_failure = NM_VPN_PLUGIN_FAILURE_LOGIN_FAILED;
-		again = FALSE;
+			again = FALSE;
+		}
+
+		g_free (auth);
 	}
 
 out:
@@ -492,28 +598,27 @@ nm_openvpn_connect_timer_cb (gpointer data)
 {
 	NMOpenvpnPlugin *plugin = NM_OPENVPN_PLUGIN (data);
 	NMOpenvpnPluginPrivate *priv = NM_OPENVPN_PLUGIN_GET_PRIVATE (plugin);
-	struct sockaddr_in     serv_addr;
-	gboolean               connected = FALSE;
-	gint                   socket_fd = -1;
 	NMOpenvpnPluginIOData *io_data = priv->io_data;
+	struct sockaddr_un remote = { 0 };
+	int fd;
 
 	priv->connect_count++;
 
 	/* open socket and start listener */
-	socket_fd = socket (PF_INET, SOCK_STREAM, IPPROTO_TCP);
-	if (socket_fd < 0)
-		return FALSE;
+	fd = socket (AF_UNIX, SOCK_STREAM, 0);
+	if (fd < 0) {
+		g_warning ("Could not create management socket");
+		nm_vpn_plugin_failure (NM_VPN_PLUGIN (plugin), NM_VPN_PLUGIN_FAILURE_CONNECT_FAILED);
+		nm_vpn_plugin_set_state (NM_VPN_PLUGIN (plugin), NM_VPN_SERVICE_STATE_STOPPED);
+		goto out;
+	}
 
-	serv_addr.sin_family = AF_INET;
-	if (inet_pton (AF_INET, "127.0.0.1", &(serv_addr.sin_addr)) <= 0)
-		g_warning ("%s: could not convert 127.0.0.1", __func__);
-	serv_addr.sin_port = htons (1194);
- 
-	connected = (connect (socket_fd, (struct sockaddr *) &serv_addr, sizeof (serv_addr)) == 0);
-	if (!connected) {
-		close (socket_fd);
+	remote.sun_family = AF_UNIX;
+	g_strlcpy (remote.sun_path, priv->mgt_path, sizeof (remote.sun_path));
+	if (connect (fd, (struct sockaddr *) &remote, sizeof (remote)) != 0) {
+		close (fd);
 		if (priv->connect_count <= 30)
-			return TRUE;
+			return G_SOURCE_CONTINUE;
 
 		priv->connect_timer = 0;
 
@@ -521,22 +626,17 @@ nm_openvpn_connect_timer_cb (gpointer data)
 		nm_vpn_plugin_failure (NM_VPN_PLUGIN (plugin), NM_VPN_PLUGIN_FAILURE_CONNECT_FAILED);
 		nm_vpn_plugin_set_state (NM_VPN_PLUGIN (plugin), NM_VPN_SERVICE_STATE_STOPPED);
 	} else {
-		GIOChannel *openvpn_socket_channel;
-		guint openvpn_socket_channel_eventid;
-
-		openvpn_socket_channel = g_io_channel_unix_new (socket_fd);
-		openvpn_socket_channel_eventid = g_io_add_watch (openvpn_socket_channel,
-		                                                 G_IO_IN,
-		                                                 nm_openvpn_socket_data_cb,
-		                                                 plugin);
-
-		g_io_channel_set_encoding (openvpn_socket_channel, NULL, NULL);
-		io_data->socket_channel = openvpn_socket_channel;
-		io_data->socket_channel_eventid = openvpn_socket_channel_eventid;
+		io_data->socket_channel = g_io_channel_unix_new (fd);
+		g_io_channel_set_encoding (io_data->socket_channel, NULL, NULL);
+		io_data->socket_channel_eventid = g_io_add_watch (io_data->socket_channel,
+		                                                  G_IO_IN,
+		                                                  nm_openvpn_socket_data_cb,
+		                                                  plugin);
 	}
 
+out:
 	priv->connect_timer = 0;
-	return FALSE;
+	return G_SOURCE_REMOVE;
 }
 
 static void
@@ -725,10 +825,53 @@ add_cert_args (GPtrArray *args, NMSettingVPN *s_vpn)
 	}
 }
 
+static void
+update_io_data_from_vpn_setting (NMOpenvpnPluginIOData *io_data,
+                                 NMSettingVPN *s_vpn,
+                                 const char *default_username)
+{
+	const char *tmp;
+
+	if (default_username) {
+		g_free (io_data->default_username);
+		io_data->default_username = g_strdup (default_username);
+	}
+
+	g_free (io_data->username);
+	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_USERNAME);
+	io_data->username = tmp ? g_strdup (tmp) : NULL;
+
+	if (io_data->password) {
+		memset (io_data->password, 0, strlen (io_data->password));
+		g_free (io_data->password);
+	}
+	tmp = nm_setting_vpn_get_secret (s_vpn, NM_OPENVPN_KEY_PASSWORD);
+	io_data->password = tmp ? g_strdup (tmp) : NULL;
+
+	if (io_data->priv_key_pass) {
+		memset (io_data->priv_key_pass, 0, strlen (io_data->priv_key_pass));
+		g_free (io_data->priv_key_pass);
+	}
+	tmp = nm_setting_vpn_get_secret (s_vpn, NM_OPENVPN_KEY_CERTPASS);
+	io_data->priv_key_pass = tmp ? g_strdup (tmp) : NULL;
+
+	g_free (io_data->proxy_username);
+	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_HTTP_PROXY_USERNAME);
+	io_data->proxy_username = tmp ? g_strdup (tmp) : NULL;
+
+	if (io_data->proxy_password) {
+		memset (io_data->proxy_password, 0, strlen (io_data->proxy_password));
+		g_free (io_data->proxy_password);
+	}
+	tmp = nm_setting_vpn_get_secret (s_vpn, NM_OPENVPN_KEY_HTTP_PROXY_PASSWORD);
+	io_data->proxy_password = tmp ? g_strdup (tmp) : NULL;
+}
+
 static gboolean
 nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
                                  NMSettingVPN *s_vpn,
                                  const char *default_username,
+                                 const char *uuid,
                                  GError **error)
 {
 	NMOpenvpnPluginPrivate *priv = NM_OPENVPN_PLUGIN_GET_PRIVATE (plugin);
@@ -736,6 +879,9 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 	GPtrArray *args;
 	GSource *openvpn_watch;
 	GPid pid;
+	gboolean dev_type_is_tap;
+	char *stmp;
+	const char *defport, *proto_tcp;
 
 	/* Find openvpn */
 	openvpn_binary = nm_find_openvpn ();
@@ -760,9 +906,8 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
  		}
  	}
 
-	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_CONNECTION_TYPE);
-	connection_type = validate_connection_type (tmp);
-	if (!connection_type) {
+	connection_type = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_CONNECTION_TYPE);
+	if (!validate_connection_type (connection_type)) {
 		g_set_error (error,
 		             NM_VPN_PLUGIN_ERROR,
 		             NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
@@ -774,15 +919,71 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 	args = g_ptr_array_new ();
 	add_openvpn_arg (args, openvpn_binary);
 
+	defport = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_PORT);
+	if (defport && !defport[0])
+		defport = NULL;
+
+	proto_tcp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_PROTO_TCP);
+	if (proto_tcp && !proto_tcp[0])
+		proto_tcp = NULL;
+
 	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_REMOTE);
-	if (tmp && strlen (tmp)) {
-		char *tok;
-		while ((tok = strsep((char**)&tmp, " ,")) != NULL) {
-			if (strlen(tok)) {
+	if (tmp && *tmp) {
+		char *tok, *port, *proto;
+		char *tmp_dup = strdup (tmp);
+		while ((tok = strsep (&tmp_dup, " \t,")) != NULL) {
+			if (*tok) {
+				port = strchr (tok, ':');
+				proto = port ? strchr (port + 1, ':') : NULL;
+				if (port)
+					*port++ = '\0';
+				if (proto)
+					*proto++ = '\0';
+
 				add_openvpn_arg (args, "--remote");
 				add_openvpn_arg (args, tok);
+				if (port) {
+					if (!add_openvpn_arg_int (args, port)) {
+						g_set_error (error,
+						             NM_VPN_PLUGIN_ERROR,
+						             NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
+						             _("Invalid port number '%s'."), port);
+						free_openvpn_args (args);
+						g_free (tmp_dup);
+						return FALSE;
+					}
+				} else if (defport) {
+					if (!add_openvpn_arg_int (args, defport)) {
+						g_set_error (error,
+						             NM_VPN_PLUGIN_ERROR,
+						             NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
+						             _("Invalid port number '%s'."),
+						             defport);
+						free_openvpn_args (args);
+						return FALSE;
+					}
+				} else
+					add_openvpn_arg (args, "1194"); /* default IANA port */
+
+				if (proto) {
+					if (!strcmp (proto, "udp") || !strcmp (proto, "tcp"))
+						add_openvpn_arg (args, proto);
+					else {
+						g_set_error (error,
+						             NM_VPN_PLUGIN_ERROR,
+						             NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
+						             _("Invalid proto '%s'."), proto);
+						free_openvpn_args (args);
+						g_free (tmp_dup);
+						return FALSE;
+					}
+				} else if (proto_tcp && !strcmp (proto_tcp, "yes"))
+					add_openvpn_arg (args, "tcp");
+				else
+					add_openvpn_arg (args, "udp");
 			}
 		}
+		g_free (tmp_dup);
 	}
 
 	/* Remote random */
@@ -814,6 +1015,7 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 				         NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
 				         _("Invalid proxy type '%s'."),
 				         tmp);
+			free_openvpn_args (args);
 			return FALSE;
 		}
 	}
@@ -824,38 +1026,29 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 
 	add_openvpn_arg (args, "--nobind");
 
-	/* Device, either tun or tap */
+	/* Device and device type, defaults to tun */
+	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_DEV);
+	tmp2 = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_DEV_TYPE);
+	tmp3 = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_TAP_DEV);
 	add_openvpn_arg (args, "--dev");
-	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_TAP_DEV);
-	if (tmp && !strcmp (tmp, "yes"))
+	if (tmp) {
+		add_openvpn_arg (args, tmp);
+		dev_type_is_tap = g_str_has_prefix (tmp, "tap");
+	} else if (tmp2) {
+		add_openvpn_arg (args, tmp2);
+	} else if (tmp3 && !strcmp (tmp3, "yes")) {
 		add_openvpn_arg (args, "tap");
-	else
-		add_openvpn_arg (args, "tun");
-
-	/* Protocol, either tcp or udp */
-	add_openvpn_arg (args, "--proto");
-	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_PROTO_TCP);
-	if (tmp && !strcmp (tmp, "yes"))
-		add_openvpn_arg (args, "tcp-client");
-	else
-		add_openvpn_arg (args, "udp");
-
-	/* Port */
-	add_openvpn_arg (args, "--port");
-	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_PORT);
-	if (tmp && strlen (tmp)) {
-		if (!add_openvpn_arg_int (args, tmp)) {
-			g_set_error (error,
-			             NM_VPN_PLUGIN_ERROR,
-			             NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
-			             _("Invalid port number '%s'."),
-			             tmp);
-			free_openvpn_args (args);
-			return FALSE;
-		}
+		dev_type_is_tap = TRUE;
 	} else {
-		/* Default to IANA assigned port 1194 */
-		add_openvpn_arg (args, "1194");
+		add_openvpn_arg (args, "tun");
+		dev_type_is_tap = FALSE;
+	}
+
+	/* Add '--dev-type' if the type was explicitly set */
+	if (tmp2) {
+		add_openvpn_arg (args, "--dev-type");
+		add_openvpn_arg (args, tmp2);
+		dev_type_is_tap = (g_strcmp0 (tmp2, "tap") == 0);
 	}
 
 	/* Cipher */
@@ -865,6 +1058,21 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 		add_openvpn_arg (args, tmp);
 	}
 
+	/* Keysize */
+	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_KEYSIZE);
+	if (tmp && strlen (tmp)) {
+		add_openvpn_arg (args, "--keysize");
+		if (!add_openvpn_arg_int (args, tmp)) {
+			g_set_error (error,
+			             NM_VPN_PLUGIN_ERROR,
+			             NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
+			             _("Invalid keysize '%s'."),
+			             tmp);
+			free_openvpn_args (args);
+			return FALSE;
+		}
+	}
+
 	/* Auth */
 	if (auth) {
 		add_openvpn_arg (args, "--auth");
@@ -890,6 +1098,13 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
                 add_openvpn_arg (args, tmp);
 	}
 
+	/* remote-cert-tls */
+	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_REMOTE_CERT_TLS);
+	if (tmp && strlen (tmp)) {
+                add_openvpn_arg (args, "--remote-cert-tls");
+                add_openvpn_arg (args, tmp);
+	}
+
 	/* Reneg seconds */
 	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_RENEG_SECONDS);
 	if (tmp && strlen (tmp)) {
@@ -959,10 +1174,10 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 
 	/* Up script, called when connection has been established or has been restarted */
 	add_openvpn_arg (args, "--up");
-	if (debug)
-		add_openvpn_arg (args, NM_OPENVPN_HELPER_PATH " --helper-debug");
-	else
-		add_openvpn_arg (args, NM_OPENVPN_HELPER_PATH);
+	stmp = g_strdup_printf ("%s%s %s --", NM_OPENVPN_HELPER_PATH, debug ? " --helper-debug" : "",
+	                        dev_type_is_tap ? "--tap" : "--tun");
+	add_openvpn_arg (args, stmp);
+	g_free (stmp);
 	add_openvpn_arg (args, "--up-restart");
 
 	/* Keep key and tun if restart is needed */
@@ -971,11 +1186,20 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 
 	/* Management socket for localhost access to supply username and password */
 	add_openvpn_arg (args, "--management");
-	add_openvpn_arg (args, "127.0.0.1");
-	/* with have nobind, thus 1194 should be free, it is the IANA assigned port */
-	add_openvpn_arg (args, "1194");
+	g_warn_if_fail (priv->mgt_path == NULL);
+	g_free (priv->mgt_path);
+	priv->mgt_path = g_strdup_printf (LOCALSTATEDIR "/run/NetworkManager/nm-openvpn-%s", uuid);
+	add_openvpn_arg (args, priv->mgt_path);
+	add_openvpn_arg (args, "unix");
+	add_openvpn_arg (args, "--management-client-user");
+	add_openvpn_arg (args, "root");
+	add_openvpn_arg (args, "--management-client-group");
+	add_openvpn_arg (args, "root");
+
 	/* Query on the management socket for user/pass */
 	add_openvpn_arg (args, "--management-query-passwords");
+	add_openvpn_arg (args, "--auth-retry");
+	add_openvpn_arg (args, "interact");
 
 	/* do not let openvpn setup routes or addresses, NM will handle it */
 	add_openvpn_arg (args, "--route-noexec");
@@ -1051,6 +1275,13 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 
 	g_ptr_array_add (args, NULL);
 
+	if (debug) {
+		char *cmd = g_strjoinv (" ", (char **) args->pdata);
+
+		g_message ("EXEC: '%s'", cmd);
+		g_free (cmd);
+	}
+
 	if (!g_spawn_async (NULL, (char **) args->pdata, NULL,
 	                    G_SPAWN_DO_NOT_REAP_CHILD, NULL, NULL, &pid, error)) {
 		free_openvpn_args (args);
@@ -1077,25 +1308,7 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 	    || nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_HTTP_PROXY_USERNAME)) {
 
 		priv->io_data = g_malloc0 (sizeof (NMOpenvpnPluginIOData));
-
-		tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_USERNAME);
-		priv->io_data->username = tmp ? g_strdup (tmp) : NULL;
-		/* Use the default username if it wasn't overridden by the user */
-		if (!priv->io_data->username && default_username)
-			priv->io_data->username = g_strdup (default_username);
-
-		tmp = nm_setting_vpn_get_secret (s_vpn, NM_OPENVPN_KEY_PASSWORD);
-		priv->io_data->password = tmp ? g_strdup (tmp) : NULL;
-
-		tmp = nm_setting_vpn_get_secret (s_vpn, NM_OPENVPN_KEY_CERTPASS);
-		priv->io_data->priv_key_pass = tmp ? g_strdup (tmp) : NULL;
-
-		tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_HTTP_PROXY_USERNAME);
-		priv->io_data->proxy_username = tmp ? g_strdup (tmp) : NULL;
-
-		tmp = nm_setting_vpn_get_secret (s_vpn, NM_OPENVPN_KEY_HTTP_PROXY_PASSWORD);
-		priv->io_data->proxy_password = tmp ? g_strdup (tmp) : NULL;
-
+		update_io_data_from_vpn_setting (priv->io_data, s_vpn, default_username);
 		nm_openvpn_schedule_connect_timer (plugin);
 	}
 
@@ -1106,6 +1319,7 @@ static const char *
 check_need_secrets (NMSettingVPN *s_vpn, gboolean *need_secrets)
 {
 	const char *tmp, *key, *ctype;
+	NMSettingSecretFlags secret_flags = NM_SETTING_SECRET_FLAG_NONE;
 
 	g_return_val_if_fail (s_vpn != NULL, FALSE);
 	g_return_val_if_fail (need_secrets != NULL, FALSE);
@@ -1123,12 +1337,22 @@ check_need_secrets (NMSettingVPN *s_vpn, gboolean *need_secrets)
 		if (is_encrypted (key) && !nm_setting_vpn_get_secret (s_vpn, NM_OPENVPN_KEY_CERTPASS))
 			*need_secrets = TRUE;
 
-		if (!nm_setting_vpn_get_secret (s_vpn, NM_OPENVPN_KEY_PASSWORD))
+		if (!nm_setting_vpn_get_secret (s_vpn, NM_OPENVPN_KEY_PASSWORD)) {
 			*need_secrets = TRUE;
+			if (nm_setting_get_secret_flags (NM_SETTING (s_vpn), NM_OPENVPN_KEY_PASSWORD, &secret_flags, NULL)) {
+				if (secret_flags & NM_SETTING_SECRET_FLAG_NOT_REQUIRED)
+					*need_secrets = FALSE;
+			}
+		}
 	} else if (!strcmp (ctype, NM_OPENVPN_CONTYPE_PASSWORD)) {
 		/* Will require a password */
-		if (!nm_setting_vpn_get_secret (s_vpn, NM_OPENVPN_KEY_PASSWORD))
+		if (!nm_setting_vpn_get_secret (s_vpn, NM_OPENVPN_KEY_PASSWORD)) {
 			*need_secrets = TRUE;
+			if (nm_setting_get_secret_flags (NM_SETTING (s_vpn), NM_OPENVPN_KEY_PASSWORD, &secret_flags, NULL)) {
+				if (secret_flags & NM_SETTING_SECRET_FLAG_NOT_REQUIRED)
+					*need_secrets = FALSE;
+			}
+		}
 	} else if (!strcmp (ctype, NM_OPENVPN_CONTYPE_TLS)) {
 		/* May require private key password */
 		key = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_KEY);
@@ -1147,65 +1371,71 @@ check_need_secrets (NMSettingVPN *s_vpn, gboolean *need_secrets)
 }
 
 static gboolean
-real_connect (NMVPNPlugin   *plugin,
-              NMConnection  *connection,
-              GError       **error)
+_connect_common (NMVPNPlugin   *plugin,
+                 NMConnection  *connection,
+                 GHashTable    *details,
+                 GError       **error)
 {
 	NMSettingVPN *s_vpn;
 	const char *connection_type;
 	const char *user_name;
-	gboolean need_secrets;
 
-	s_vpn = NM_SETTING_VPN (nm_connection_get_setting (connection, NM_TYPE_SETTING_VPN));
+	s_vpn = nm_connection_get_setting_vpn (connection);
 	if (!s_vpn) {
-		g_set_error (error,
-		             NM_VPN_PLUGIN_ERROR,
-		             NM_VPN_PLUGIN_ERROR_CONNECTION_INVALID,
-		             "%s",
-		             _("Could not process the request because the VPN connection settings were invalid."));
+		g_set_error_literal (error,
+		                     NM_VPN_PLUGIN_ERROR,
+		                     NM_VPN_PLUGIN_ERROR_CONNECTION_INVALID,
+		                     _("Could not process the request because the VPN connection settings were invalid."));
 		return FALSE;
 	}
 
-	/* Check if we need secrets and validate the connection type */
-	connection_type = check_need_secrets (s_vpn, &need_secrets);
-	if (!connection_type) {
-		g_set_error (error,
-		             NM_VPN_PLUGIN_ERROR,
-		             NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
-		             "%s",
-		             _("Invalid connection type."));
+	connection_type = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_CONNECTION_TYPE);
+	if (!validate_connection_type (connection_type)) {
+		g_set_error_literal (error,
+		                     NM_VPN_PLUGIN_ERROR,
+		                     NM_VPN_PLUGIN_ERROR_CONNECTION_INVALID,
+		                     _("Could not process the request because the openvpn connection type was invalid."));
 		return FALSE;
 	}
 
-	user_name = nm_setting_vpn_get_user_name (s_vpn);
-
-	/* Need a username for any password-based connection types */
-	if (   !strcmp (connection_type, NM_OPENVPN_CONTYPE_PASSWORD_TLS)
-	    || !strcmp (connection_type, NM_OPENVPN_CONTYPE_PASSWORD)) {
-		if (!user_name && !nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_USERNAME)) {
-			g_set_error (error,
-			             NM_VPN_PLUGIN_ERROR,
-			             NM_VPN_PLUGIN_ERROR_CONNECTION_INVALID,
-			             "%s",
-			             _("Could not process the request because no username was provided."));
-			return FALSE;
-		}
-	}
-
 	/* Validate the properties */
 	if (!nm_openvpn_properties_validate (s_vpn, error))
 		return FALSE;
 
 	/* Validate secrets */
-	if (need_secrets) {
-		if (!nm_openvpn_secrets_validate (s_vpn, error))
-			return FALSE;
-	}
-
-	/* Finally try to start OpenVPN */
-	if (!nm_openvpn_start_openvpn_binary (NM_OPENVPN_PLUGIN (plugin), s_vpn, user_name, error))
+	if (!nm_openvpn_secrets_validate (s_vpn, error))
 		return FALSE;
 
+	/* Finally try to start OpenVPN */
+	user_name = nm_setting_vpn_get_user_name (s_vpn);
+	if (!nm_openvpn_start_openvpn_binary (NM_OPENVPN_PLUGIN (plugin),
+	                                      s_vpn,
+	                                      user_name,
+	                                      nm_connection_get_uuid (connection),
+	                                      error))
+		return FALSE;
+
+	return TRUE;
+}
+
+static gboolean
+real_connect (NMVPNPlugin   *plugin,
+              NMConnection  *connection,
+              GError       **error)
+{
+	return _connect_common (plugin, connection, NULL, error);
+}
+
+static gboolean
+real_connect_interactive (NMVPNPlugin   *plugin,
+                          NMConnection  *connection,
+                          GHashTable    *details,
+                          GError       **error)
+{
+	if (!_connect_common (plugin, connection, details, error))
+		return FALSE;
+
+	NM_OPENVPN_PLUGIN_GET_PRIVATE (plugin)->interactive = TRUE;
 	return TRUE;
 }
 
@@ -1227,23 +1457,21 @@ real_need_secrets (NMVPNPlugin *plugin,
 		nm_connection_dump (connection);
 	}
 
-	s_vpn = NM_SETTING_VPN (nm_connection_get_setting (connection, NM_TYPE_SETTING_VPN));
+	s_vpn = nm_connection_get_setting_vpn (connection);
 	if (!s_vpn) {
-		g_set_error (error,
-		             NM_VPN_PLUGIN_ERROR,
-		             NM_VPN_PLUGIN_ERROR_CONNECTION_INVALID,
-		             "%s",
-		             _("Could not process the request because the VPN connection settings were invalid."));
+		g_set_error_literal (error,
+		                     NM_VPN_PLUGIN_ERROR,
+		                     NM_VPN_PLUGIN_ERROR_CONNECTION_INVALID,
+		                     _("Could not process the request because the VPN connection settings were invalid."));
 		return FALSE;
 	}
 
 	connection_type = check_need_secrets (s_vpn, &need_secrets);
 	if (!connection_type) {
-		g_set_error (error,
-		             NM_VPN_PLUGIN_ERROR,
-		             NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
-		             "%s",
-		             _("Invalid connection type."));
+		g_set_error_literal (error,
+		                     NM_VPN_PLUGIN_ERROR,
+		                     NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
+		                     _("Invalid connection type."));
 		return FALSE;
 	}
 
@@ -1253,6 +1481,50 @@ real_need_secrets (NMVPNPlugin *plugin,
 	return need_secrets;
 }
 
+static gboolean
+real_new_secrets (NMVPNPlugin *plugin,
+                  NMConnection *connection,
+                  GError **error)
+{
+	NMOpenvpnPluginPrivate *priv = NM_OPENVPN_PLUGIN_GET_PRIVATE (plugin);
+	NMSettingVPN *s_vpn;
+	const char *message = NULL;
+	char **hints = NULL;
+
+	s_vpn = nm_connection_get_setting_vpn (connection);
+	if (!s_vpn) {
+		g_set_error_literal (error,
+		                     NM_VPN_PLUGIN_ERROR,
+		                     NM_VPN_PLUGIN_ERROR_CONNECTION_INVALID,
+		                     _("Could not process the request because the VPN connection settings were invalid."));
+		return FALSE;
+	}
+
+	if (debug)
+		g_message ("VPN received new secrets; sending to management interface");
+
+	update_io_data_from_vpn_setting (priv->io_data, s_vpn, NULL);
+
+	g_warn_if_fail (priv->io_data->pending_auth);
+	if (!handle_auth (priv->io_data, priv->io_data->pending_auth, &message, &hints)) {
+		g_set_error_literal (error,
+		                     NM_VPN_PLUGIN_ERROR,
+		                     NM_VPN_PLUGIN_ERROR_GENERAL,
+		                     _("Unhandled pending authentication."));
+		return FALSE;
+	}
+
+	/* Request new secrets if we need any */
+	if (message) {
+		if (debug)
+			g_message ("Requesting new secrets: '%s'", message);
+		nm_vpn_plugin_secrets_required (plugin, message, (const char **) hints);
+	}
+	if (hints)
+		g_free (hints);  /* elements are 'const' */
+	return TRUE;
+}
+
 static gboolean
 ensure_killed (gpointer data)
 {
@@ -1298,8 +1570,10 @@ nm_openvpn_plugin_class_init (NMOpenvpnPluginClass *plugin_class)
 
 	/* virtual methods */
 	parent_class->connect      = real_connect;
+	parent_class->connect_interactive = real_connect_interactive;
 	parent_class->need_secrets = real_need_secrets;
 	parent_class->disconnect   = real_disconnect;
+	parent_class->new_secrets  = real_new_secrets;
 }
 
 static void
@@ -1382,17 +1656,26 @@ main (int argc, char *argv[])
 		{NULL}
 	};
 
+#if !GLIB_CHECK_VERSION (2, 35, 0)
 	g_type_init ();
+#endif
+
+	/* locale will be set according to environment LC_* variables */
+	setlocale (LC_ALL, "");
+
+	bindtextdomain (GETTEXT_PACKAGE, NM_OPENVPN_LOCALEDIR);
+	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
+	textdomain (GETTEXT_PACKAGE);
 
 	/* Parse options */
-	opt_ctx = g_option_context_new ("");
-	g_option_context_set_translation_domain (opt_ctx, "UTF-8");
+	opt_ctx = g_option_context_new (NULL);
+	g_option_context_set_translation_domain (opt_ctx, GETTEXT_PACKAGE);
 	g_option_context_set_ignore_unknown_options (opt_ctx, FALSE);
 	g_option_context_set_help_enabled (opt_ctx, TRUE);
 	g_option_context_add_main_entries (opt_ctx, options, NULL);
 
 	g_option_context_set_summary (opt_ctx,
-		_("nm-vpnc-service provides integrated OpenVPN capability to NetworkManager."));
+		_("nm-openvpn-service provides integrated OpenVPN capability to NetworkManager."));
 
 	g_option_context_parse (opt_ctx, &argc, &argv, NULL);
 	g_option_context_free (opt_ctx);
@@ -1403,7 +1686,8 @@ main (int argc, char *argv[])
 	if (debug)
 		g_message ("nm-openvpn-service (version " DIST_VERSION ") starting...");
 
-	if (system ("/sbin/modprobe tun") == -1)
+	if (   !g_file_test ("/sys/class/misc/tun", G_FILE_TEST_EXISTS)
+	    && (system ("/sbin/modprobe tun") == -1))
 		exit (EXIT_FAILURE);
 
 	plugin = nm_openvpn_plugin_new ();
